/*
 * Copyright (C) 2008 Kikini Limited and contributors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.kikini.util.cmdline;

import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;

/**
 * Annotates a field or a property setter method as a command line argument. The
 * {@link CommandLine} class can then be used to update the property, given a
 * command-line string array.
 * <p>
 * A typical use of this annotation looks like this:
 * <p>
 * <blockquote><tt>
 *   <b>&#064;Argument</b><br>
 *   private boolean flag;
 * </tt></blockquote>
 * <p>
 * The basic example above enables a user to control the value of the {@code
 * flag} field by specifying "<tt>--flag</tt> " on the command line. If it is
 * specified, the field will be set to {@code true}, if not, it will be left at
 * its default value. Users can also control the value directly, by adding (for
 * example) " <tt>--flag=false</tt>" to the command line.
 * <p>
 * This value-assignment style works for other types of fields and properties,
 * not just booleans. For example:
 * <p>
 * <blockquote><tt>
 *   <b>&#064;Argument</b><br>
 *   private String param;
 * </tt></blockquote>
 * <p>
 * The example above declares an argument named "param", which can be specified
 * on the command line as "<tt>--param=value</tt>". Arguments can be any
 * primitive type, as well as any type that has a public, single-argument String
 * constructor, such as {@link java.io.File File}. You can annotate any type
 * that has an associated {@link java.beans.PropertyEditor PropertyEditor}
 * implementation; if you write a custom {@code PropertyEditor}, you need to
 * ensure that it can be found by a {@link java.beans.PropertyEditorManager
 * PropertyEditorManager}.
 * <p>
 * A field can also be any {@link java.util.Collection Collection} type, which
 * allows multiple values to be specified on the command line. This is
 * particularly useful for the "nameless" parameter type:
 * <p>
 * <blockquote><tt>
 *   <b>&#064;Argument(nameless = true)</b><br>
 *   private List&lt;Path&gt; paths;
 * </tt></blockquote>
 * <p>
 * Note that collections are not supported for properties, only fields.
 * 
 * @author Jon Vincent
 */
@Target(value = { ElementType.FIELD, ElementType.METHOD })
@Retention(RetentionPolicy.RUNTIME)
public @interface Argument {

    /** Specifies whether this argument is required. */
    boolean required() default false;

    /**
     * Specifies whether this parameter is nameless — that is, whether this is
     * the "default" argument. If this attribute is {@code true}, the {@code
     * name}, {@code shortcut}, and {@code caseSensitive} attributes are
     * ignored.
     */
    boolean nameless() default false;

    /**
     * Specifies the name of the parameter. If this attribute is not specified
     * (or is empty), the name of the property or field is used.
     */
    String name() default "";

    /** The shortcut character for the parameter. */
    char shortcut() default 0;

    /**
     * Specifies whether this parameter's name (and shortcut, if any) are
     * case-sensitive.
     */
    boolean caseSensitive() default true;

    /**
     * Specifies the resource key for the argument description. If this
     * attribute is not specified (or is empty), the argument has no
     * description.
     */
    String description() default "";

}