/*
 * Copyright (C) 2008 Kikini Limited and contributors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.kikini.util.cmdline;

import java.beans.Introspector;
import java.beans.PropertyEditor;
import java.beans.PropertyEditorManager;
import java.beans.PropertyEditorSupport;
import java.io.PrintStream;
import java.lang.reflect.AccessibleObject;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Yet another command line parser. This one uses annotations on the class to
 * specify which bean properties or fields should be set by command-line
 * arguments.
 * <p>
 * This class expects arguments to be specified in one of three ways:
 * <p>
 * <ul>
 * <li>As a standard argument: <tt>program arg arg</tt></li>
 * <li>As a parameter: <tt>program --param1 --param2=value</tt></li>
 * <li>As a flag: <tt>program -fdE -a=value -n</tt></li>
 * </ul>
 * 
 * @author Jon Vincent
 */
public final class CommandLine {

    private static final Pattern LINE_BREAK = Pattern.compile("[ \\t\\x0B\\f\\r]*\\n?[ \\t\\x0B\\f\\r]*");
    private static final int MAX_USAGE_WIDTH = 80;

    private final String[] args;
    private final BitSet applied;

    /** Creates a new instance. */
    public CommandLine(String[] args) {
        if (args == null) throw new NullPointerException();
        this.args = args;
        this.applied = new BitSet(args.length);
    }

    /**
     * Applies this command line to the specified object, based on its
     * {@link Argument @CommandLineArgument} annotations.
     * 
     * @param target
     *        the target object
     * @throws ArgumentFormatException
     *         if an illegal argument is encountered
     * @throws MissingArgumentException
     *         if a required argument is not specified
     */
    public void applyTo(Object target) throws ArgumentFormatException, MissingArgumentException {
        // Get the properties for this object
        List<Property> props = getProperties(target.getClass());
        final int propCount = props.size();
        BitSet propsApplied = new BitSet(propCount);

        // Iterate through the command-line arguments
        for (int argIndex = 0, argCount = args.length; argIndex < argCount; ++argIndex) {
            // Try to find a matching property
            String arg = args[argIndex];
            for (int propIndex = 0; propIndex < propCount; ++propIndex) {
                // Match the argument against the property's pattern
                Property prop = props.get(propIndex);
                Matcher matcher = prop.pattern.matcher(arg);
                if (matcher.matches()) try {
                    // Check for a null assignment
                    String value = matcher.group(1);
                    if (value == null) {
                        if (prop.requiresValue) {
                            throw ArgumentFormatException.forMissingValue(arg, prop.getType());
                        }

                        // If the property doesn't require a value, it's a
                        // boolean, and should be treated as true
                        value = "true";
                    }

                    // Set the value
                    prop.set(target, value);
                    propsApplied.set(propIndex);
                    applied.set(argIndex);
                    break;
                } catch (InvocationTargetException e) {
                    Throwable cause = e.getCause();
                    if (cause instanceof Error) throw (Error)cause;
                    if (cause instanceof RuntimeException) throw (RuntimeException)cause;
                    throw ArgumentFormatException.forInvalidValue(arg, matcher.group(1), prop.getType(), cause);
                } catch (Exception e) {
                    throw ArgumentFormatException.forInvalidValue(arg, matcher.group(1), prop.getType(), e);
                }
            }
        }

        // Now iterate through the properties, looking for required properties
        for (int propIndex = propsApplied.nextClearBit(0); propIndex < propCount; propIndex = propsApplied.nextClearBit(propIndex + 1)) {
            Property prop = props.get(propIndex);
            if (prop.anno.required()) {
                throw MissingArgumentException.forTarget(prop.name);
            }
        }
    }

    /**
     * Checks that all of the command-line arguments have been applied to one or
     * more objects.
     * 
     * @throws UnrecognizedArgumentException
     *         if any arguments were not applied
     */
    public void verifyNoMoreArguments() throws UnrecognizedArgumentException {
        // Get the first unused argument
        int index = applied.nextClearBit(0);
        if (index < args.length) throw UnrecognizedArgumentException.forArgument(args[index]);
    }

    /**
     * Writes parameter information for the specified class, based on its
     * {@link Argument @CommandLineArgument} annotations.
     * <p>
     * The format is:
     * 
     * <pre>
     *   -p=ARG [--param]  : this option has a shortcut and takes a value
     *   -o [--option]     : a long description for a boolean which shows how the text
     *                       can span multiple lines
     *   --another-option  : boolean that has no shortcut character
     * </pre>
     * 
     * @param out
     *        the output writer
     * @param clazz
     *        the target class
     * @param resources
     *        the localized resources to use, or {@code null} to output the
     *        description strings directly
     */
    public static void usage(PrintStream out, Class<?> clazz, ResourceBundle resources) {
        // Find all of the properties for the class
        List<Property> props = getProperties(clazz);

        // Build the label list
        List<String> labels = new ArrayList<String>(props.size());
        int maxLength = 0;
        for (Property prop : props) {
            // Construct the format string
            String format = (prop.shortcut != 0 ? "-%3$s%2$s [--%1$s]" : "--%s%s");
            String assignment = (prop.requiresValue ? "=ARG" : "");

            // Add the label to the list
            String label = String.format(format, prop.name, assignment, prop.shortcut);
            labels.add(label);
            maxLength = Math.max(maxLength, label.length());
        }

        // Check the length
        int indent = 6 + maxLength;
        if (indent >= MAX_USAGE_WIDTH - 1) throw new AssertionError("Parameter names are too long");

        // Write each property description
        String formatLabelDesc = "  %-" + maxLength + "s  : %s%n";
        for (int i = 0, l = props.size(); i < l; ++i) {
            // Skip nameless properties
            Property prop = props.get(i);
            if (prop.name.length() == 0) continue;

            // Get the description for the property
            String desc = prop.anno.description();
            if (desc.length() != 0 && resources != null) {
                desc = resources.getString(desc);
            }

            // Split the description into multiple lines
            desc = breakText(desc, indent, MAX_USAGE_WIDTH - 1);

            // Write the property description
            String format = (desc.length() != 0 ? formatLabelDesc : "  %s%n");
            out.printf(format, labels.get(i), desc);
        }
    }

    // package-private for test only
    static List<Property> getProperties(Class<?> clazz) {
        List<AccessibleObject> needsAccessibility = new ArrayList<AccessibleObject>();
        List<Property> list = new ArrayList<Property>();
        while (clazz != Object.class) {
            // Iterate through all of the class fields
            for (Field field : clazz.getDeclaredFields()) {
                // Does the field have an annotation?
                Argument anno = field.getAnnotation(Argument.class);
                if (anno == null) continue;

                // Add the property
                addProperty(new Property(anno, field), clazz, list);
                needsAccessibility.add(field);
            }

            // Iterate through all of the class methods
            for (Method method : clazz.getDeclaredMethods()) {
                // Does the method have an annotation?
                Argument anno = method.getAnnotation(Argument.class);
                if (anno == null) continue;

                // Is this a property setter?
                Class<?>[] params = method.getParameterTypes();
                if (params.length != 1 || method.getReturnType() != void.class) {
                    String message = "CommandLineArgument annotation on invalid method [" + method.getName() + "]";
                    throw new AssertionError(message);
                }

                // Add the property
                addProperty(new Property(anno, method), clazz, list);
                needsAccessibility.add(method);
            }

            // Move on to the superclass
            clazz = clazz.getSuperclass();
        }

        // Mark all of the properties as accessible
        AccessibleObject[] array = new AccessibleObject[needsAccessibility.size()];
        AccessibleObject.setAccessible(needsAccessibility.toArray(array), true);
        return list;
    }

    private static void addProperty(Property prop, Class<?> clazz, List<Property> list) {
        // Does the property already exist in the list?
        int index = Collections.binarySearch(list, prop);
        if (index >= 0) {
            String format;
            if (prop.name.length() == 0) {
                format = "Nameless command-line annotation in [%2$s] conflicts with existing annotation";
            } else {
                format = "Command-line annotation [%1$s] in [%2$s] conflicts with existing annotation";
            }

            throw new AssertionError(String.format(format, prop.name, clazz.getName()));
        }

        // Note that this probably makes more sense as a SortedSet
        // -- but the problem is that our comparison function is not
        // consistent with equals (and in fact it's impossible to
        // obey the contract of equals() and hashCode() for our sort
        // function)
        list.add(~index, prop);
    }

    /**
     * Breaks a string to split it across multiple lines. The algorithm used is
     * very simple, and just looks for whitespace to split the text. It doesn't
     * support hyphenation, and assumes that all characters are the same width.
     * For this reason, it's probably only suitable for command-line display.
     * <p>
     * This method does support embedded newline characters, and replaces them
     * with the system line separator (as well as indenting them).
     * 
     * @param text
     *        the source string
     * @param indent
     *        the number of characters to indent secondary lines
     * @param maxLength
     *        the maximum line length, in characters
     * @return the updated text
     */
    public static String breakText(String text, int indent, int maxLength) {
        // Remove the indent size from the max length
        maxLength -= indent;

        // Create the control objects
        String lineBreak = System.getProperty("line.separator");
        int indentLength = lineBreak.length() + indent;
        StringBuilder builder = new StringBuilder(text.length() + indentLength * (text.length() / maxLength));
        Pattern line = Pattern.compile("\\G.{0," + (maxLength - 1) + "}\\S(?!\\S)");
        Matcher matcher = line.matcher(text);

        // Create the indent array
        char[] indentArray = new char[indent];
        Arrays.fill(indentArray, ' ');

        // Find the matches
        int offset = 0, breakCount = 0;
        while (offset < text.length()) {
            // Write any line breaks
            for (; breakCount > 0; --breakCount) {
                builder.append(lineBreak).append(indentArray);
            }

            // Try to find a line break
            if (matcher.find(offset)) {
                // Write the text
                builder.append(matcher.group());
                offset = matcher.end();
            } else {
                // Write as much as we can
                builder.append(text, offset, offset + maxLength);
                offset += maxLength;
                breakCount = 1;
                continue;
            }

            // Eat any whitespace
            matcher.usePattern(LINE_BREAK);
            while (true) {
                matcher.find(offset);
                int newOffset = matcher.end();
                if (newOffset == offset) break;
                offset = newOffset;
                ++breakCount;
            }

            // Reset the search
            matcher.usePattern(line);
        }

        // Success
        return builder.toString();
    }

    // package-private for test only
    static final class Property implements Comparable<Property> {

        private final Argument anno;

        private final PropertyEditor editor;
        private final boolean requiresValue;
        private final boolean useCollection;
        private final AccessibleObject ao;

        final String name;
        final char shortcut;

        private final boolean caseSensitive;
        private final String nameIgnoreCase;
        private final char shortcutIgnoreCase;

        final Pattern pattern;

        // package-private for test
        Property(Argument anno, AccessibleObject ao) {
            // Save the annotation
            this.anno = anno;

            // Get the property name
            if (anno.nameless()) {
                // Use the "default" name parameters
                name = "";
                nameIgnoreCase = "";
                shortcut = 0;
                shortcutIgnoreCase = 0;
                caseSensitive = false;
            } else {
                // Copy the name parameters from the annotation
                String name = anno.name();
                shortcut = anno.shortcut();
                caseSensitive = anno.caseSensitive();

                // Do we need to update the name?
                if (name.length() == 0) {
                    if (ao instanceof Field) {
                        name = ((Field)ao).getName();
                    } else if (ao instanceof Method) {
                        name = ((Method)ao).getName();
                        if (name.startsWith("set")) {
                            name = Introspector.decapitalize(name.substring(3));
                        }
                    }
                }

                // Save the name
                this.name = name;
                nameIgnoreCase = name.toUpperCase(Locale.ENGLISH).toLowerCase(Locale.ENGLISH);
                String s = String.valueOf(shortcut).toUpperCase(Locale.ENGLISH).toLowerCase(Locale.ENGLISH);
                this.shortcutIgnoreCase = s.charAt(0);
            }

            // Get the property type
            Class<?> propertyClass;
            if (ao instanceof Field) {
                propertyClass = ((Field)ao).getType();
            } else if (ao instanceof Method) {
                propertyClass = ((Method)ao).getParameterTypes()[0];
            } else {
                // We should only hit this from test code
                propertyClass = String.class;
            }

            // Is the property a collection?
            useCollection = (Collection.class.isAssignableFrom(propertyClass) && (ao instanceof Field));
            if (useCollection) {
                // Try to detect the collection type
                Class<?> elementClass = getCollectionType(((Field)ao).getGenericType());
                if (elementClass == null) {
                    String format = "Could not determine type of [%s] in [%s]";
                    throw new AssertionError(String.format(format, name, ((Field)ao).getDeclaringClass().getName()));
                }

                // Save the class
                propertyClass = elementClass;
            }

            // The PropertyEditorManager doesn't support the "wrapped"
            // primitives
            if (propertyClass == Boolean.class) propertyClass = boolean.class;
            if (propertyClass == Byte.class) propertyClass = byte.class;
            if (propertyClass == Short.class) propertyClass = short.class;
            if (propertyClass == Character.class) propertyClass = char.class;
            if (propertyClass == Integer.class) propertyClass = int.class;
            if (propertyClass == Long.class) propertyClass = long.class;
            if (propertyClass == Float.class) propertyClass = float.class;
            if (propertyClass == Double.class) propertyClass = double.class;

            // Find an editor for the property type
            PropertyEditor editor = PropertyEditorManager.findEditor(propertyClass);
            if (editor == null && propertyClass == char.class) editor = new CharEditor();
            if (editor == null) editor = StringConstructibleEditor.forClass(propertyClass);
            if (editor == null) {
                // Better to check this here than when setting the value, since
                // this will catch bugs even if the application code doesn't
                // test a given property
                throw new AssertionError("Could not find property editor for [" + propertyClass + "]");
            }

            // Set up the type-specific parameters
            this.editor = editor;
            requiresValue = (propertyClass != boolean.class);
            pattern = createPattern();

            // Save the accessible object
            this.ao = ao;
        }

        private static Class<?> getCollectionType(Type type) {
            if (!(type instanceof ParameterizedType)) return null;
            Type[] types = ((ParameterizedType)type).getActualTypeArguments();
            if (types.length != 1 || !(types[0] instanceof Class<?>)) return null;
            return (Class<?>)types[0];
        }

        private Pattern createPattern() {
            // The nameless pattern just matches all non-hyphenated strings
            if (name.length() == 0) return Pattern.compile("([^-].*)?");

            // Construct the format string
            String format = (shortcut != 0 ? "-(?:-%s|%s)" : "--%s") + "(?:=(.*))?";

            // Create the pattern
            String name = Pattern.quote(this.name);
            String shortcut = Pattern.quote(String.valueOf(this.shortcut));
            int flags = (caseSensitive ? 0 : Pattern.CASE_INSENSITIVE);
            return Pattern.compile(String.format(format, name, shortcut), flags);
        }

        private void set(Object target, String value) throws Exception, InvocationTargetException {
            // Get the parsed value
            editor.setAsText(value);
            Object parsed = editor.getValue();

            try {
                if (useCollection) {
                    Field field = (Field)ao;

                    // Do we need to create a new collection?
                    @SuppressWarnings("unchecked")
                    Collection<Object> collection = (Collection)field.get(target);
                    if (collection == null) {
                        collection = createCollection(field.getType());
                        field.set(target, collection);
                    }

                    // Add the value to the collection
                    collection.add(parsed);
                } else if (ao instanceof Field) {
                    // Set the value directly
                    ((Field)ao).set(target, parsed);
                } else {
                    // Call the setter method
                    ((Method)ao).invoke(target, parsed);
                }
            } catch (IllegalAccessException e) {
                // this shouldn't happen since we called SetAccessible()
                throw new AssertionError(e);
            }
        }

        private static Collection<Object> createCollection(Class<?> clazz) {
            // If the class isn't an interface, create it directly
            if (!clazz.isInterface()) {
                try {
                    @SuppressWarnings("unchecked")
                    Collection<Object> collection = (Collection)clazz.newInstance();
                    return collection;
                } catch (IllegalAccessException e) {
                    throw new AssertionError("Could not access constructor for collection [" + clazz.getName() + "]");
                } catch (InstantiationException e) {
                    throw new AssertionError("Instantiation failed for collection [" + clazz.getName() + "]");
                }
            } else if (clazz.isAssignableFrom(Set.class)) {
                return new HashSet<Object>();
            } else if (clazz.isAssignableFrom(List.class)) {
                return new ArrayList<Object>();
            } else if (clazz.isAssignableFrom(Queue.class)) {
                return new LinkedList<Object>();
            } else if (clazz.isAssignableFrom(SortedSet.class)) {
                return new TreeSet<Object>();
            }

            // We don't know how to create this collection
            throw new AssertionError("Don't know how to create implementation for collection [" + clazz.getName() + "]");
        }

        @Override
        public int compareTo(Property o) {
            // Check whether this comparison should be case-sensitive
            boolean caseSensitive = (this.caseSensitive && o.caseSensitive);

            // Does this item have a shortcut?
            if (shortcutIgnoreCase != 0) {
                if (o.shortcutIgnoreCase == 0) return -1;
                if (shortcutIgnoreCase < o.shortcutIgnoreCase) return -1;
                if (shortcutIgnoreCase > o.shortcutIgnoreCase) return 1;
                if (caseSensitive) {
                    if (shortcut < o.shortcut) return -1;
                    if (shortcut > o.shortcut) return 1;
                }

                // They're the same shortcut
                return 0;
            }

            // Does the other item have a shortcut?
            if (o.shortcutIgnoreCase != 0) return 1;

            // Compare the names
            for (int i = 0, thisLength = nameIgnoreCase.length(), thatLength = o.nameIgnoreCase.length();; ++i) {
                // Shorter strings are "less than" longer strings
                if (i == thatLength) return (i != thisLength ? 1 : 0);
                if (i == thisLength) return -1;

                // Compare characters
                char thisChar = nameIgnoreCase.charAt(i), thatChar = o.nameIgnoreCase.charAt(i);
                if (thisChar < thatChar) return -1;
                if (thisChar > thatChar) return 1;
                if (caseSensitive) {
                    thisChar = name.charAt(i);
                    thatChar = o.name.charAt(i);
                    if (thisChar < thatChar) return -1;
                    if (thisChar > thatChar) return 1;
                }
            }
        }

        @Override
        public String toString() {
            StringBuilder builder = new StringBuilder();
            if (shortcut != 0) {
                builder.append('-').append(shortcut).append(' ').append('[');
            }

            if (name.length() != 0) {
                builder.append('-').append('-');
                builder.append(name);
                if (shortcut != 0) builder.append(']');
            } else {
                builder.append("[nameless]");
            }

            builder.append(' ').append('(');
            builder.append(getType());
            builder.append(')');

            return builder.toString();
        }

        private Class<?> getType() {
            return ((ao instanceof Field) ? ((Field)ao).getType() : ((Method)ao).getParameterTypes()[0]);
        }

    }

    private static final class CharEditor extends PropertyEditorSupport {

        @Override
        public void setAsText(String text) throws IllegalArgumentException {
            if (text.length() != 1) throw new IllegalArgumentException(text);
            setValue(Character.valueOf(text.charAt(0)));
        }

    }

    private static final class StringConstructibleEditor extends PropertyEditorSupport {

        private final Constructor<?> constructor;

        private StringConstructibleEditor(Constructor<?> constructor) {
            this.constructor = constructor;
        }

        private static StringConstructibleEditor forClass(Class<?> clazz) {
            try {
                // Try to create an editor for this type
                if (Modifier.isAbstract(clazz.getModifiers())) return null;
                Constructor<?> constructor = clazz.getConstructor(String.class);
                constructor.setAccessible(true);
                return new StringConstructibleEditor(constructor);
            } catch (NoSuchMethodException e) {
                // No big deal, there's just no editor for this type
                return null;
            }
        }

        @Override
        public void setAsText(String text) throws IllegalArgumentException {
            try {
                setValue(constructor.newInstance(text));
            } catch (InvocationTargetException e) {
                Throwable cause = e.getCause();
                if (cause instanceof RuntimeException) throw (RuntimeException)cause;
                throw new IllegalArgumentException(cause);
            } catch (IllegalAccessException e) {
                // We explicitly set the constructor to accessible
                throw new AssertionError(e);
            } catch (InstantiationException e) {
                // Typically this means the class is abstract (which we checked
                // for)
                throw new AssertionError(e);
            }
        }

    }

}